module Parser where

type Parser a b = [a] -> [(b, [a])]

succeed :: b -> Parser a b
succeed v inp = [(v,inp)]

parsefail :: Parser a b
parsefail inp = []

satisfy :: (a -> Bool) -> Parser a a
satisfy p []                 = parsefail []
satisfy p (x:xs) | p x       = succeed x xs
                 | otherwise = parsefail xs

literal :: (Eq a) => a -> Parser a a
literal x = satisfy (== x)

parseAlt :: Parser a b -> Parser a b -> Parser a b
parseAlt p1 p2 inp = (p1 inp) ++ (p2 inp)

parseOpt :: Parser a b -> b -> Parser a b
parseOpt p v = p `parseAlt` (succeed v)

parseThen :: Parser a b -> Parser a c -> Parser a (b,c)
parseThen p1 p2 inp = [((v1,v2), out2) | (v1,out1) <- p1 inp,
                                          (v2,out2) <- p2 out1]

					  
parseUsing :: Parser a b -> (b -> c) -> Parser a c
parseUsing p f inp = [(f v, out) | (v,out) <- p inp]

parseMany :: Parser a b -> Parser a [b]
parseMany p = ((p `parseThen` (parseMany p)) `parseUsing` (\(x,xs) -> x:xs)) `parseAlt` (succeed []) 

parseSome :: Parser a b -> Parser a [b]
parseSome p = (p `parseThen` (parseMany p)) `parseUsing` (\(x,xs) -> x:xs)

parseNumber :: Parser Char [Char]
parseWord :: Parser Char [Char]
parseCapital :: Parser Char [Char]
parseIdentifier :: Parser Char [Char]
parseNumber = (((satisfy minus) `parseThen` parseNumberInner) 
                `parseUsing` (\ (x,xs) -> x:xs)) `parseAlt` parseNumberInner
              where minus x = x == '-'
parseNumberInner = (parseSome (satisfy digit))
              where digit x = (x `elem` ['0'..'9'])
parseWord = parseSome (satisfy letter)
            where letter x = (x `elem` ['a'..'z']) || (x `elem` ['A'..'Z'])
parseCapital = (satisfy capital) `parseUsing` (\ c -> [c])
             where capital x = (x `elem` ['A'..'Z'])
parseIdentifier = ((satisfy letter) `parseThen` (parseMany (satisfy part))) `parseUsing` (\(x,xs) -> x:xs)
                where part x = (letter x) || (digit x) || (punctuation x)
                      letter x = (x `elem` ['a'..'z']) || (x `elem` ['A'..'Z'])
		      digit x = (x `elem` ['0'..'9'])
		      punctuation x = (x `elem` ['-','_']) 

parseString :: (Eq a) => [a] -> Parser a [a]
parseString [] = succeed []
parseString (x:xs) = ((literal x) `parseThen` (parseString xs)) `parseUsing` (\(x,xs) -> x:xs)

parseXThen :: Parser a b -> Parser a c -> Parser a c
parseThenX :: Parser a b -> Parser a c -> Parser a b
parseXThen p1 p2 = (p1 `parseThen` p2) `parseUsing` snd
parseThenX p1 p2 = (p1 `parseThen` p2) `parseUsing` fst

parseReturn :: Parser a b -> c -> Parser a c
parseReturn p v = p `parseUsing` (\x -> v)

parseAny :: (a -> Parser b c) -> [a] -> Parser b c
parseAny p = foldr (parseAlt . p) parsefail

nibble :: Parser Char a -> Parser Char a
nibble p = white `parseXThen` p `parseThenX` white
           where white = parseMany (parseAny literal " \t\n")

parseSymbol :: [Char] -> Parser Char [Char]
parseSymbol = nibble . parseString

-- everything is generic as far as here and can be exported.
-- modules that import this one will probably need:
--   define tags for the various flavours of syntactic object
--   define a type, Token = (Tag,[Char])
--   define matchToken and a low level lexer that are used to 
--      recognise individual tokens.

--matchToken :: Parser Char [Char] -> Tag -> Parser Char Token
--matchToken p t inp = [((t,xs), out) | (xs, out) <- p inp ]
--                     where x = head inp

--lowLex :: [(Parser Char [Char], Tag)] -> Parser Char [Token]
--lowLex = parseMany . (foldr op parsefail)
--         where op (p,t) xs = (p `matchToken` t) `parseAlt` xs

--   define the actual lexer, in terms of lowLex, so that it 
--     goes tags each component string of input.
--   define strip; used to remove junk tokens from the lexed input.
--   define kind, used to recognise whether a particular token has a 
--     particular tag.
--   define lit, used to recognize particular strings with particular 
--     tags.

-- Then need to build up the particular syntax using these combinators, 
-- plus some extra functions that can convert raw parsed input into actual
-- part of the proper data structure. 

 