Automatic test generation from a Test Specification Language (TSL) file.

First of all I want to get the module line out of the way

>module TestGen (FuncSpec(..), Id(..), FuncN(..), Category(..), CatId(..), 
>                CatN(..), FTag(..), Choice(..), Chid(..), ChN(..), 
>                Property(..), Constraint(..), Condition(..), Flag(..),
>                TestFrame(..), TestFrameBody(..), testgen, prepare) where
>import Parser
>import QSort (sortLe)

This import is for ghc only (equality only defined up to 5-tuples in ghc).

>import PreludeSixTuple

These two imports might be necessary one day, when the libraries exist.

import LibReadTextIO
import LibWriteTextIO


Now, the important bit is going to be the data structure for a FuncSpec,
the basic unit of a test specification.

>type FuncSpec = (Id, FuncN, [Category])

A FuncSpec consists of its Id, a string of capital letters,

>type Id = String

its FuncN, the name of the function

>type FuncN = String

and a collection of categories. The specification gives this as a
function from category identifiers to actual categories, but I will
implement category identifiers as simply numbers, so a sequence will
suffice.

>type Category = (CatId, CatN, FTag, Choices)

Each CatId (a number) uniquely identifies its category. It should also be 
the same as the categories position in the list.

>type CatId = Int

Each category also has a name, CatN.

>type CatN = String

Finally, each category is made up of a collection of choices. Choices
are quite complex. They are made up from an identifying number and
name, plus a possible list of properties, a possible constraint and a possible
flag.

>type Choices = [Choice]

type Choice = (Chid,ChN,Property,Constraint,Flag)
type Chid = Int

>type Choice = (Chid,ChN,Property,Constraint,Flag,FTag)
>type Chid = Int
>type ChN = String
>data Property = NoPpty | Ppty [String] deriving (Eq, Text)
>data Constraint = NoCnstrnt | Cnstrnt Condition deriving (Eq, Text)
>data Condition = PropCond String 
>               | NotCond Condition
>	        | AndCond Condition Condition
>		| OrCond  Condition Condition deriving (Eq, Text)
>data Flag = NoFlag | Single | Error deriving (Eq, Text)
>data FTag  = NoFTag | FTag String deriving (Eq, Text)

There are lots of functions that need to be defined based on the
structure of choices.

Firstly, and most obviously, conditions and constraints need to be
evaluated, but it only makes sense to do so in a particular context. A
context will turn out to be a list of properties that are "in force"
at a particular time. 

>type Context = [Property]

Which is all very well, but I will usually want to work with the list
of strings rather than the list of properties.

>deProp :: [Property] -> [String]
>deProp [] = []
>deProp (NoPpty:ps) = deProp ps
>deProp ((Ppty ss):ps) = ss ++ (deProp ps)

Now I can evaluate conditions

>evalCondition :: Condition -> Context -> Bool
>evalCondition (PropCond str) ctxt  = str `elem` (deProp ctxt)
>evalCondition (NotCond c) ctxt     = not (evalCondition c ctxt)
>evalCondition (AndCond c1 c2) ctxt = (evalCondition c1 ctxt) && (evalCondition c2 ctxt)
>evalCondition (OrCond c1 c2) ctxt  = (evalCondition c1 ctxt) || (evalCondition c2 ctxt)

and then constraints in their turn, bearing in mind that a constraint
is either empty (always evaluating to true) or a wrapped up condition.

>evalConstraint :: Constraint -> Context -> Bool
>evalConstraint NoCnstrnt _ = True
>evalConstraint (Cnstrnt c) ctxt = evalCondition c ctxt

Test Frame Generation:

The body of a test frame consists of one choice from each of a function's
categories, so it can be implemented as a sequence of choices.

>type TestFrameBody = [Choice]

Eacg test frame body has some associated information. In particular
each frame has a unique identification number.

>type TestFrame = (Id, Int, TestFrameBody)

Now I can finally write the the test generation function.

The final function will be this shape:

>testgen :: FuncSpec -> [TestFrame]

But a large number of intermediate steps are needed before this
function can written.

Firstly, convert the test spec into a plain list of lists of choices.

>convertToPlainList :: FuncSpec -> [[Choice]]
>convertToPlainList (id, name, cats) = [catchoices | 
>                                        (catid, catname, cattag, catchoices) <- cats]

I need a function to distribute a list through a list of lists.

>distribute :: [a] -> [[a]] -> [[a]]
>distribute [] [] = []
>distribute (x:xs) [] = (x:[]) : (distribute xs [])
>distribute xs (y:ys) = y : ([ x:y | x <- xs ] ++ (distribute xs ys))

Now, I want to recurse through the list of lists of choices, applying
distribute as I go. A fold should do the trick. The result of this
needs to be filtered according to whether the result is a valid test
frame according to the rules.

The rules for validity are:
A test frame is either "flagged" or "unflagged".
Unflagged frames:
  1. Must be made up of exactly one choice for each category.
  2. Must not include any flagged choices.
  3. Must be internally consistent, in that every constraint is satisfied.

Flagged frames:
  1. Contains a flagged "primary" choice from amongst all of the
     categories' choices.
  2. Contains the smallest possible number of other choices required
     to satisfy the primary choice's constraints.

This version is indiscriminate.

generateAllCombinations :: [[Choice]] -> [TestFrameBody]
generateAllCombinations chss = [chs | chs <- (foldr distribute [] chss)]

This version tries to remove the invalid frames.

>generateAllCombinations :: [[Choice]] -> [TestFrameBody]
>generateAllCombinations chss = [chs | chs <- (foldr distribute [] chss),
>                                      validFrame chs (length chss)]

In order to be valid, each constraint in a test frame must evaluate
true in the context of this frame. A function is needed to extract the
context of a frame.

>extractContext :: TestFrameBody -> Context
>extractContext chs = [prop | (_,_,prop,_,_,_) <- chs]

It will also be handy to extract the list of constraints of a frame.

>extractConstraints :: TestFrameBody -> [Constraint]
>extractConstraints chs = [const | (_,_,_,const,_,_) <- chs] 

Determining the validity of a frame is now a simple matter of firstly
checking that either the frame has one choice for each category or it
is flagged and then evaluating each constraint in context, and
checking that none of them evaluate false.

>validFrame :: TestFrameBody -> Int -> Bool
>validFrame frame size = if (size == (length frame)) || (isFlagged frame)
>                        then let ctxt = extractContext frame 
>                             in and (map (\c -> (evalConstraint c ctxt)) (extractConstraints frame))
>                        else False

>isFlagged :: TestFrameBody -> Bool
>isFlagged choices = or (map (\(_,_,_,_,flag,_) -> not (flag == NoFlag)) choices) 

Given a list of flagged frames, we want to reduce it so that each
flagged choice only occurs once, if possible. This can be achieved by
scanning along the list of flagged frames, and transferring a frame to
the final list iff it contains a flagged choice that is not already
represented in the final list.

>thinFlaggedFrames :: [TestFrameBody] -> [TestFrameBody] -> [TestFrameBody]
>thinFlaggedFrames [] final = final
>thinFlaggedFrames (frame:fs) final | (newFlag frame final) = (thinFlaggedFrames fs (frame:final))
>                                   | otherwise             = thinFlaggedFrames fs final  

So, a function to recognise when a frame contains a new flag is required.
A function to extract just the flagged choices from a frame will be useful.

>extractFlags :: TestFrameBody -> [Choice]
>extractFlags chs = filter (\(_,_,_,_,flag,_) -> not (flag == NoFlag)) chs
>newFlag :: TestFrameBody -> [TestFrameBody] -> Bool
>newFlag frame frames = innerNewFlag (extractFlags frame)
>                                    (concat (map extractFlags frames))
>innerNewFlag :: [Choice] -> [Choice] -> Bool
>innerNewFlag [] _ = False
>innerNewFlag _ [] = True
>innerNewFlag (ch:chs) choices = not (ch `elem` choices) || 
>                                (innerNewFlag chs choices)

If the flagged frames are sorted into length order before going
through the thinFlaggedFrames process, then only the shortest example
frames will survive. Use a quicksort from the HBC library to order them.

>shorterFrame :: TestFrameBody -> TestFrameBody -> Bool
>shorterFrame t1 t2 = (length t1) <= (length t2)

>lengthOrder :: [TestFrameBody] -> [TestFrameBody]
>lengthOrder tfs = sortLe shorterFrame tfs


So, overall, the set of test frames is made up of the unflagged ones,
plus the remains of the thinned out flagged ones.

generate :: [[Choice]] -> [TestFrameBody]
generate choices = generateAllCombinations choices

This version attempts to thin down the flagged frames, so that each
flagged choice will only appear once, if possible. It will succeed as
long as all the flagged choices are unique.

>generate :: [[Choice]] -> [TestFrameBody]
>generate choices = let (flagged,unflagged) = partition isFlagged 
>                                                (generateAllCombinations choices)
>                   in (thinFlaggedFrames (lengthOrder flagged) []) ++ unflagged

generate produces a list of TestFrameBodies. These need to be processed
to add the function id and to number the individual cases.

>numberFrames :: Id -> Int -> [TestFrameBody] -> [TestFrame]
>numberFrames _ _ [] = []
>numberFrames i n (f:fs) = (i,n,f):(numberFrames i (n+1) fs)

It is now a simple matter to define the overall test generation function:

>testgen spec@(id,name,cats) = numberFrames id 1 (generate (convertToPlainList spec))



Input parsing

I have imported the parser module to provide the basic parsing
combinators.

However, a number of functions need to be specially defined for the TSL.

Tokenising the input

Firstly, I need a type to store complete Tokens (a piece of input with
an appropriate tag.  In turn, this means I need to define all of the
tags needed for this language.
  Junk - spaces, tabs, comments (a line begining with #)
  Eoln - newlines
  ReservedWord - Fixed parts of the syntax
  Capital - sequence of capital letters (used for FuncSpec ID)
  Separator - * char, used to identify beginning of a new choice
  Word - an identifier (cannot start with a digit). Used as names for functions, 
         categories and choices (where they are recombined into sentences
  Number - sequence of digits; used as unique id for categories
  FrameTag - anything between '{' and '}'

>type Token = (Tag,[Char])

>data Tag = Junk | Eoln | ReservedWord | Capital | Separator | 
>           Word | Number | FrameTag                          deriving (Text,Eq)

Now that the Tokens are clearly defined, I need some functions to
convert raw input into them. Firstly, matchToken can be used to add
tags to arbitrary parsed strings.

>matchToken :: Parser Char [Char] -> Tag -> Parser Char Token
>matchToken p t inp = [((t,xs), out) | (xs, out) <- p inp ]
>                     where x = head inp

lowLex is given a long list of simple parsers plus suitable tags, and
produces a parser that will tag complete input strings.

>lowLex :: [(Parser Char [Char], Tag)] -> Parser Char [Token]
>lowLex = parseMany . (foldr op parsefail)
>         where op (p,t) xs = (p `matchToken` t) `parseAlt` xs

lexer is simply written in terms of lowLex. It actually provides the
correspondence between the string components of the language and their
tags.

>lexer :: Parser Char [Token]
>lexer = lowLex [(parseSome (parseAny literal " \t"), Junk ),
>             (parseString ['\n'], Eoln),
>             (parseAny parseString ["Function:", "EndFunction;", "[property", "[if", 
>                                    "[single]","[error]","not(",")and(",")or("], ReservedWord),
>             (parseString "*", Separator),
>             (parseCapital , Capital),
> 	      (parseIdentifier , Word),
>             (parseAny parseString ["(",",",")","]"], ReservedWord),
>	      (parseNumber, Number),
>             (parseFrTag, FrameTag),
>             (parseComment, Junk)]

All of these are defined (in Parser.hs) apart from parseFrTag and parseComment.

>parseFrTag :: Parser Char [Char]
>parseFrTag = (((literal '@') 
>             `parseXThen` (parseMany (satisfy notCloseFTag)))
>             `parseThenX` (literal '@'))
>             where notCloseFTag c = (c /= '@')

>parseComment :: Parser Char [Char]
>parseComment = (literal '#') `parseXThen` (parseMany (satisfy notEoln)) 
>               `parseThenX` (parseString ['\n'])
>               where notEoln c = (c /= '\n')

Some tokens are of no further use and can safely be stripped out.

>strip :: [Token] -> [Token]
>strip = filter (\(t,_) -> (t /= Junk))

The whole process of converting an input string into the equivalent
tokenised form can be wrapped up in a single simple function.

>tokenise s = strip (fst (head (lexer s)))

Parsing the tokenised input string

Before I begin, I will need to be able to identify particular kinds of
token, regardless of the actual value.

>kind :: Tag -> Parser Token [Char]
>kind t = (satisfy (\(tg,_) -> (t==tg))) `parseUsing` snd

Also, there are lots of different ReservedWords and I need to be able
to distinguish them.

>lit :: [Char] -> Parser Token [Char]
>lit xs = literal (ReservedWord,xs) `parseUsing` snd

Now, I can finally get down to the actual parser. This is simply a
transposition of the BNF syntax rules using all of the corresponding
parser combinators.

>aFuncSpec = ((kind Capital) `parseThenX` (lit "Function:") 
>                     `parseThen` (kind Word) `parseThenX` (kind Eoln) 
>            `parseThen` (parseMany aCatSpec) 
>                `parseThenX` (lit "EndFunction;") `parseThenX` (kind Eoln)) `parseUsing` enTriple

>enTriple :: ((a,b),c) -> (a,b,c)
>enTriple ((x,y),z) = (x,y,z)

>enQuad :: ((a,b,c),d) -> (a,b,c,d)
>enQuad ((a,b,c),d) = (a,b,c,d)

>aCatSpec :: Parser Token Category
>aCatSpec = ((((((kind Number) `parseUsing` toNumber) 
>                 `parseThen` (parseMany (parseAny kind [Word,Number,Capital]) `parseUsing` unwords)) 
>                 `parseThen` aFTag) `parseUsing` enTriple)
>                 `parseThenX` (kind Eoln)
>                 `parseThen` (parseMany aChoice))
>           `parseUsing` enQuad
>toNumber :: String -> Int
>toNumber s = foldl1 (\n m -> ((n*10)+m)) (map charToInt s)
>             where charToInt c = ord c - ord '0'

(unwords is a standard function to convert a list of words into a
sentence of space separated words.)

>aChoice :: Parser Token Choice
>aChoice = ((((((kind Separator) 
>          `parseXThen` (parseMany (parseAny kind [Word,Number,Capital]) `parseUsing` unwords))
>          `parseThen` aProperty)
>	   `parseThen` aConstraint `parseUsing` enTriple) 
>	   `parseThen` aFlag `parseUsing` enQuad)
>          `parseThen` aFTag `parseUsing` enChoice) `parseThenX` (kind Eoln)

>aProperty :: Parser Token Property
>aProperty = (((lit "[property") 
>            `parseXThen` (((kind Word) `parseUsing` (\s -> [s])) 
>                 `parseAlt` (((kind Word) 
>                            `parseThen` (parseSome ((lit ",") `parseXThen` (kind Word)))) 
>                            `parseUsing` (\ (x,xs) -> x:xs) )
>            `parseThenX` (lit "]"))) `parseOpt` [])
>            `parseUsing` enProp
>enProp :: [String] -> Property
>enProp [] = NoPpty
>enProp ss = Ppty ss

>aConstraint :: Parser Token Constraint
>aConstraint = (((lit "[if") `parseXThen` aCondition `parseThenX` (lit "]"))
>              `parseUsing` (\c -> Cnstrnt c)) `parseOpt` NoCnstrnt

>aCondition :: Parser Token Condition
>aCondition = ((kind Word) `parseUsing` (\s -> PropCond s))  `parseAlt`
>             (((lit "not(") `parseXThen` aCondition `parseThenX` (lit ")")) 
>                                              `parseUsing` (\c -> NotCond c)) `parseAlt`
>             (((lit "(") `parseXThen` aCondition `parseThenX` (lit ")and(")
>                         `parseThen` aCondition `parseThenX` (lit ")"))
>                                              `parseUsing` (\ (c1,c2) -> AndCond c1 c2)) `parseAlt`
>             (((lit "(") `parseXThen` aCondition `parseThenX` (lit ")or(")
>                         `parseThen` aCondition `parseThenX` (lit ")"))
>                                              `parseUsing` (\ (c1,c2) -> OrCond c1 c2))

>aFlag :: Parser Token Flag
>aFlag = (((lit "[single]") `parseAlt` (lit "[error]") `parseOpt` []))
>        `parseUsing` enFlag
>enFlag :: String -> Flag
>enFlag "" = NoFlag
>enFlag "[single]" = Single
>enFlag "[error]" = Error

>aFTag :: Parser Token FTag
>aFTag = ((kind FrameTag) `parseOpt` [])
>        `parseUsing` enFTag
>enFTag :: String -> FTag
>enFTag "" = NoFTag
>enFTag s  = FTag s

enChoice is used to make the final conversion into a Choice. It adds a
dummy Chid of 0 to every choice. This can be replaced later by the
correct category number.

>enChoice :: ((ChN,Property,Constraint,Flag),FTag) -> Choice
>enChoice ((a,b,c,d),e) = (0,a,b,c,d,e)

I want a function that can be used to do the complete conversion from
raw String to parsed FuncSpec.

>prepare :: String -> FuncSpec
>prepare ss = numberChoices (fst (head (aFuncSpec (tokenise ss))))

Now, the Choices all need to be given their proper identifying
identification numbers. This is simply a matter of going through the
categories and applying the category number to each choice. At the
same time, any category tags can be combined with the choice tags.

>numberChoices :: FuncSpec -> FuncSpec
>numberACat :: Category -> Category
>renumberAChoice ::  Int -> FTag -> Choice -> Choice
>combineTags :: FTag -> FTag -> FTag
>renumberAChoice n t (_,a,b,c,d,e) = (n,a,b,c,d,(combineTags t e))
>numberACat (num,name,tag,chs) = (num,name,tag, map (renumberAChoice num tag) chs)
>numberChoices (id,name,cats) = (id,name, map numberACat cats)
>combineTags NoFTag t = t
>combineTags t NoFTag = t
>combineTags (FTag s1) (FTag s2) = FTag (s1 ++ "\n\t" ++ s2)

